@@ -52,23 +52,23 @@ indent"if [ -d \""CHROOT_HOOKDIR"\" ]; then\n" \
 indent"    for fn in \""CHROOT_HOOKDIR"/"prefix"\"[0-9][0-9]* ; do\n" \
 indent"        case \"$fn\" in\n" \
 indent"            \""CHROOT_HOOKDIR"/"prefix"\"'[0-9][0-9]*')\n" \
-indent"                echo \"W: no hooks of type "prefix" found -- ignoring\"\n" \
+indent"                log.d \"no hooks of type "prefix" found -- ignoring\"\n" \
 indent"                ;;\n" \
 indent"            *~)\n" \
-indent"                echo \"W: skipping an editor backup file $fn\"\n" \
+indent"                log.w \"skipping an editor backup file $fn\"\n" \
 indent"                ;;\n" \
 indent"            *)\n" \
 indent"                if [ -x \"$fn\" ]; then\n" \
-indent"                    echo \"I: user script $fn starting\"\n" \
+indent"                    log.i \"user script $fn starting\"\n" \
 indent"                    \""CHROOT_HOOKDIR"/$(basename \"$fn\")\"\n" \
-indent"                    echo \"I: user script $fn finished\"\n" \
+indent"                    log.i \"user script $fn finished\"\n" \
 indent"                else\n" \
 indent"                    if [ -f \"$fn\" ]; then\n" \
 indent"                        filetype=$(basename \"$fn\")\n" \
-indent"                        echo \"W: execute priv not set on file $filetype, not executing.\"\n" \
+indent"                        log.w \"execute priv not set on file $filetype, not executing.\"\n" \
 indent"                    else\n" \
 indent"                        # Should it reach here ? This case should be caught in the above case.\n" \
-indent"                        echo \"W: no hooks of type ${prefix} found -- internal error in logic\"\n" \
+indent"                        log.w \"no hooks of type ${prefix} found -- internal error in logic\"\n" \
 indent"                    fi\n" \
 indent"                fi\n" \
 indent"                ;;\n" \
@@ -177,10 +177,10 @@ static FILE* create_script(const char* mountpoint, const char* relative_path)
       *c = 0;
       if (mkdir(s, 0777) && errno != EEXIST)
 	{
-	  fprintf(stderr,
-		  "E: Could not create directory '%s': %s\n",
-		  s,
-		  strerror(errno));
+	  log_printf(log_error,
+              "Could not create directory '%s': %s",
+              s,
+              strerror(errno));
 	  goto fail;
 	}
       *c = '/';
@@ -230,8 +230,8 @@ static int copy_file_contents_to_temp(const char* orig,
   ret=copy_file(orig, temppath);
   if (ret == -1)
     {
-      fprintf(stderr, "E: Copy file error in %s to %s\n",
-	      orig, temppath);
+      log_printf(log_error, "Copy file error in %s to %s",
+          orig, temppath);
       goto out;
     }
 out:
@@ -247,9 +247,9 @@ static int copy_file_contents_in_temp(FILE *f,
   int trailing_slash = targetdir[strlen(targetdir)-1] == '/';
   const char* sep = trailing_slash ? "" : "/";
   fprintf(f,
-	  "echo \"I: copying %s%s%s from temporary location\"\n"
+	  "log.i \"copying %s%s%s from temporary location\"\n"
 	  "mkdir -p %s\n"
-	  "cp $BUILDDIR/%s %s%s%s || echo \"E: Copy failed\"\n",
+	  "cp $BUILDDIR/%s %s%s%s || log.e \"Copy failed\"\n",
 	  targetdir, sep, targetname,
 	  targetdir,
 	  tempname,
@@ -270,10 +270,10 @@ static int copy_file_contents_through_temp(FILE* f,
 		 "%s/input",
 		 tempdir))
     {
-      fprintf(stderr,
-	      "E: failed to allocate string for '%s/input': %s\n",
-	      tempdir,
-	      strerror(errno));
+      log_printf(log_error,
+	  "failed to allocate string for '%s/input': %s",
+	  tempdir,
+	  strerror(errno));
       ret = 1;
       goto out;
     }
@@ -282,20 +282,20 @@ static int copy_file_contents_through_temp(FILE* f,
 		 "input/%s",
 		 file_basename))
     {
-      fprintf(stderr,
-	      "E: failed to allocate string for 'input/%s': %s\n",
-	      file_basename,
-	      strerror(errno));
+      log_printf(log_error,
+	  "failed to allocate string for 'input/%s': %s",
+	  file_basename,
+	  strerror(errno));
       ret = 1;
       goto out;
     }
 
   if (mkdir(tempinput, 0777) && errno != EEXIST)
     {
-      fprintf(stderr,
-	      "E: failed to create directory '%s': %s\n",
-	      tempinput,
-	      strerror(errno));
+      log_printf(log_error,
+	  "failed to create directory '%s': %s",
+	  tempinput,
+	  strerror(errno));
       ret = 1;
       goto out;
     }
@@ -328,29 +328,29 @@ static int copy_hookdir(const char *hookdir, const char *tmp)
 		 "%s/hooks",
 		 tmp))
     {
-      fprintf(stderr,
-	      "Error allocating string for '%s/hooks': %s\n",
-	      tmp,
-	      strerror(errno));
+      log_printf(log_error,
+	  "Error allocating string for '%s/hooks': %s",
+	  tmp,
+	  strerror(errno));
       ret = 1;
       goto out;
     }
 
   if (mkdir(hookstmp,0777))
     {
-      fprintf(stderr,
-	      "Error creating directory for hooks: %s\n",
-	      strerror(errno));
+      log_printf(log_error,
+	  "Error creating directory for hooks: %s",
+	  strerror(errno));
       ret = 1;
       goto out;
     }
 
   if (hd == NULL)
     {
-      fprintf(stderr,
-	      "Error copying hooks from '%s': %s\n",
-	      hookdir,
-	      strerror(errno));
+      log_printf(log_error,
+	  "Error copying hooks from '%s': %s",
+	  hookdir,
+	  strerror(errno));
       ret = 1;
       goto out;
     }
@@ -365,20 +365,20 @@ static int copy_hookdir(const char *hookdir, const char *tmp)
 		     hookdir,
 		     dirp->d_name))
 	{
-	  fprintf(stderr,
-		  "Error allocating string for '%s/%s': %s\n",
-		  hookdir,
-		  dirp->d_name,
-		  strerror(errno));
+	  log_printf(log_error,
+	      "Error allocating string for '%s/%s': %s",
+	      hookdir,
+	      dirp->d_name,
+	      strerror(errno));
 	  ret = 1;
 	  goto out;
 	}
       if (0>stat(src, &st))
 	{
-	  fprintf(stderr,
-		  "Error calling stat '%s': %s\n",
-		  src,
-		  strerror(errno));
+	  log_printf(log_error,
+	      "Error calling stat '%s': %s",
+	      src,
+	      strerror(errno));
 	  free(src);
 	  ret = 1;
 	  goto out;
@@ -468,7 +468,7 @@ static int fork_qemu(const char* hda, const char* hdb, const struct pbuilderconf
 	  if (0<waitpid(child, &status, WNOHANG))
 	    {
 	      /* child has exited */
-	      printf("\n   -> qemu exited unexpectedly: %d\n", status);
+	      log_printf(log_error, "qemu exited unexpectedly: %d", status);
 	      break;
 	    }
 	  FD_SET(sp[0],&readfds);
@@ -490,19 +490,19 @@ static int fork_qemu(const char* hda, const char* hdb, const struct pbuilderconf
 				       qemu_keyword, strlen(qemu_keyword)))!=0)
 		    {
 		      exit_code = atoi(matchptr + strlen(qemu_keyword));
-		      printf("\n  -> received termination message from inside qemu with exit-code %i, killing child process (qemu:%i)\n",
-			     exit_code,
-			     child);
+		      log_printf(log_info, "received termination message from inside qemu with exit-code %i, killing child process (qemu:%i)",
+			  exit_code,
+			  child);
 
 		      assert(child != 0);assert(child > 0);
 
 		      if (!kill(child, SIGTERM))
-			printf("   -> successfully killed qemu\n");
+			log_printf(log_info, "successfully killed qemu");
 		      else
-			perror("   -> failed to kill qemu? :");
+			log_perror("failed to kill qemu?");
 		      if (-1==waitpid(child, &status, 0))
 			{
-			  perror("qemubuilder: waitpid");
+			  log_perror("qemubuilder: waitpid");
 			}
 		      break;
 		    }
@@ -511,7 +511,7 @@ static int fork_qemu(const char* hda, const char* hdb, const struct pbuilderconf
 	    }
 	  else
 	    {
-	      perror("select");
+	      log_perror("select");
 	      break;
 	    }
 	}
@@ -521,10 +521,12 @@ static int fork_qemu(const char* hda, const char* hdb, const struct pbuilderconf
       /* this is the child process */
       const char* qemu = qemu_arch_qemu(pc->arch);
       const char* machine = qemu_arch_qemumachine(pc->arch);
+      const char* cpu = qemu_arch_cpu(pc->arch);
       const char* hda_format = format_for_image(hda);
       const char* hdb_format = format_for_image(hdb);
       char* hda_command;
       char* hdb_command;
+      int virt;
       char* append_command;
       const char* kernel_image = pc->kernel_image;
       const char* initrd = pc->initrd;
@@ -545,12 +547,12 @@ static int fork_qemu(const char* hda, const char* hdb, const struct pbuilderconf
 	term = "dumb";
 
       if (qemu == NULL || machine == NULL) {
-	fprintf(stderr, "Your architecture %s does not seem to be supported\n", pc->arch);
+	log_printf(log_error, "Your architecture %s does not seem to be supported", pc->arch);
 	exit(1);
       }
 
       if (kernel_image == NULL || !strcmp(kernel_image, "")) {
-	fprintf(stderr, "No KERNEL_IMAGE defined in pbuilderrc\n");
+	log_printf(log_error, "No KERNEL_IMAGE defined in pbuilderrc");
 	exit(1);
       }
 
@@ -558,22 +560,26 @@ static int fork_qemu(const char* hda, const char* hdb, const struct pbuilderconf
 
       if (hda_format == NULL)
 	{
-	  fprintf(stderr, "Unknown format for disk image %s\n", hda);
+	  log_printf(log_error, "Unknown format for disk image %s", hda);
 	  exit(1);
 	}
 
       if (hdb_format == NULL)
 	{
-	  fprintf(stderr, "Unknown format for disk image %s\n", hdb);
+	  log_printf(log_error, "Unknown format for disk image %s", hdb);
 	  exit(1);
 	}
 
+      virt = !strcmp(machine, "virt");
+
       asprintf(&hda_command,
-	       "file=%s,format=%s,index=0,media=disk,cache=writeback,id=hd0",
+	       "%sfile=%s,format=%s,index=0,media=disk,cache=writeback,id=hd0",
+	       virt ? "if=none," : "",
 	       hda, hda_format);
 
       asprintf(&hdb_command,
-	       "file=%s,format=%s,index=1,media=disk,cache=writeback,id=hd1",
+	       "%sfile=%s,format=%s,index=1,media=disk,cache=writeback,id=hd1",
+	       virt ? "if=none," : "",
 	       hdb, hdb_format);
 
       /* panic < 0 means reboot immediately on panic; this will actually
@@ -594,6 +600,11 @@ static int fork_qemu(const char* hda, const char* hdb, const struct pbuilderconf
       argv[argc++]="-no-reboot"; /* halt instead of rebooting */
       argv[argc++]="-M";
       argv[argc++]=strdupa(machine);
+      if (cpu)
+	{
+	  argv[argc++]="-cpu";
+	  argv[argc++]=strdupa(cpu);
+	}
       argv[argc++]="-m";
       argv[argc++]=mem;
       if (pc->smp) {
@@ -624,38 +635,45 @@ static int fork_qemu(const char* hda, const char* hdb, const struct pbuilderconf
 	}
       argv[argc++]="-net";
       argv[argc++]="user";
-      if (!strcmp(machine, "virt")) {
-	argv[argc++]="-device";
-	argv[argc++]="virtio-scsi-device,id=scsi";
-	argv[argc++]="-device";
-	argv[argc++]="virtio-net-device,netdev=net0";
-	argv[argc++]="-device";
-	argv[argc++]="scsi-hd,drive=hd0";
-	argv[argc++]="-device";
-	argv[argc++]="scsi-hd,drive=hd1";
-	argv[argc++]="-netdev";
-	argv[argc++]="user,id=net0";
-      } else {
-	argv[argc++]="-net";
-	argv[argc++]="nic";
-      }
+      if (virt)
+	{
+	  argv[argc++]="-device";
+	  argv[argc++]="virtio-scsi-device,id=scsi";
+	  argv[argc++]="-device";
+	  argv[argc++]="virtio-net-device,netdev=net0";
+	  argv[argc++]="-device";
+	  argv[argc++]="scsi-hd,drive=hd0";
+	  argv[argc++]="-device";
+	  argv[argc++]="scsi-hd,drive=hd1";
+	  argv[argc++]="-netdev";
+	  argv[argc++]="user,id=net0";
+	}
+      else
+	{
+	  argv[argc++]="-net";
+	  argv[argc++]="nic";
+	}
       argv[argc]=NULL;
       assert(argc < MAX_ARGS);
 
-      printf("  forking qemu: ");
+      log_begin(log_info);
+      log_middle(log_info, "forking qemu:");
       for (i=0; i<argc; ++i)
 	{
-	  printf("%s ", argv[i]);
+	  if (strchr(argv[i], ' '))
+	    log_middle(log_info, " '%s'", argv[i]);
+	  else
+	    log_middle(log_info, " %s", argv[i]);
 	}
-      printf("\n");
+      log_end(log_info);
 
       execvp(argv[0], argv);
-      perror("fork_qemu");
+      log_perror("fork_qemu");
       exit (1);
     }
   else
     {
-      perror("fork");
+      log_perror("fork");
       return -1;
     }
 
@@ -698,7 +716,9 @@ static void write_first_stage(FILE *f, const struct pbuilderconfig* pc)
   fprintf(f,
 	  "#!/bin/bash\n"
 	  "echo \n"
-	  "echo ' -> qemu-pbuilder first-stage' \n"
+	  /* Can't use log.i or check LOGLEVEL in first stage, since this lives
+	   * inside the base chroot, and has not yet mounted the input disk. */
+	  "echo 'I: qemu-pbuilder first-stage' \n"
 	  "export PBUILDER_INIT_VERSION="XSTR(PBUILDER_INIT_VERSION)"\n"
 	  "export PATH='/usr/local/sbin:/usr/local/bin:/usr/sbin:/usr/bin:/sbin:/bin'\n"
 	  "stty sane\n"
@@ -723,6 +743,108 @@ static void write_first_stage(FILE *f, const struct pbuilderconfig* pc)
 	  );
 }
 
+static void write_second_stage_header(FILE *f, int debug_shell) {
+  const char *log_level;
+
+  switch (log_get_filter_level()) {
+    case log_debug: log_level = "D"; break;
+    case log_info:  log_level = "I"; break;
+    case log_warn:  log_level = "W"; break;
+    case log_error: log_level = "E"; break;
+    default:        log_level = "I"; break;
+  }
+
+  fprintf(f,
+	  "#!/bin/bash\n"
+
+	  /* Will not be auto (using log_get_use_colors, not _unresolved) */
+	  "export USECOLORS=\"%s\"\n"
+	  "export LOGLEVEL=\"%s\"\n"
+
+	  /* Bash logging taken from pbuilder */
+	  "# Log a message\n"
+	  "# message is of a format\n"
+	  "#  E: error message\n"
+	  "#  W: warning message\n"
+	  "#  I: informational message\n"
+	  "_log() {\n"
+	  "    set -u\n"
+	  "    local color=\"$1\" ; shift\n"
+	  "    local red='\\033[0;31m'\n"
+	  "    local yellow='\\033[1;33m'\n"
+	  "    local blue='\\033[0;34m'\n"
+	  "    local reset='\\033[0m'\n"
+	  "    case \"$USECOLORS\" in\n"
+	  "        yes)\n"
+	  "            printf \"${!color}${*}${reset}\\n\"\n"
+	  "            ;;\n"
+	  "        no)\n"
+	  "            printf \"${*}\\n\"\n"
+	  "            ;;\n"
+	  "        *)\n"
+	  "            printf \"malformed value of USECOLORS: [%%s]\\n\" \"$USECOLORS\" >&2\n"
+	  "            exit 1\n"
+	  "            ;;\n"
+	  "    esac\n"
+	  "    set +u\n"
+	  "}\n"
+	  "function log() {\n"
+	  "    case \"$*\" in\n"
+	  "        \"E: \"*)\n"
+	  "            _log 'red' \"$*\" >&2\n"
+	  "            ;;\n"
+	  "        \"W: \"*)\n"
+	  "            _log 'yellow' \"$*\" >&2\n"
+	  "            ;;\n"
+	  "        \"I: \"*)\n"
+	  "            _log 'reset' \"$*\"\n"
+	  "            ;;\n"
+	  "        \"D: \"*)\n"
+	  "            _log 'blue' \"$*\"\n"
+	  "            ;;\n"
+	  "        *)\n"
+	  "            echo \"malformed log message: $*\" >&2\n"
+	  "            exit 1\n"
+	  "            ;;\n"
+	  "    esac\n"
+	  "}\n"
+
+	  "log.e() {\n"
+	  "    case \"$LOGLEVEL\" in\n"
+	  "        D|I|W|E) log \"E: $*\" ;;\n"
+	  "    esac\n"
+	  "}\n"
+	  "log.w() {\n"
+	  "    case \"$LOGLEVEL\" in\n"
+	  "        D|I|W) log \"W: $*\" ;;\n"
+	  "    esac\n"
+	  "}\n"
+	  "log.i() {\n"
+	  "    case \"$LOGLEVEL\" in\n"
+	  "        D|I) log \"I: $*\" ;;\n"
+	  "    esac\n"
+	  "}\n"
+	  "log.d() {\n"
+	  "    case \"$LOGLEVEL\" in\n"
+	  "        D) log \"D: $*\" ;;\n"
+	  "    esac\n"
+	  "}\n"
+
+	  /* define function to terminate qemu */
+	  "function exit_from_qemu() {\n"
+	  "    %s\n"
+	  "    sync\n"
+	  "    sync\n"
+	  "    sleep 1s\n"		/* sleep before sending dying message */
+	  "    log.i 'qemu-pbuilder %s'\"$1\"\n"
+	  "    sleep 1s\n"
+	  "    halt -f -p\n"		/* just halt myself if possible */
+	  "}\n",
+	  log_get_use_colors() == log_use_colors_yes ? "yes" : "no",
+	  log_level,
+	  debug_shell?"log.i \"Debug shell\"; /bin/bash":"",
+	  qemu_keyword);
+}
 
 /**
  * Invoke qemu, and run the second-stage script within QEMU.
@@ -752,7 +874,7 @@ static int run_second_stage_script
   if (mkdir(pc->buildplace,0777))
     {
       /* could not create the buildplace here. */
-      perror("mkdir");
+      log_perror("mkdir");
       goto out;
     }
 
@@ -765,38 +887,25 @@ static int run_second_stage_script
   loop_mount(workblockdevicepath, pc->buildplace);
 
   f = create_script(pc->buildplace, "input/pbuilder-run");
-  fprintf(f,
-	  "#!/bin/bash\n"
-
-	  /* define function to terminate qemu */
-	  "function exit_from_qemu() {\n"
-	  "sync\n"
-	  "sync\n"
-	  "sleep 1s\n"		/* sleep before sending dying message */
-	  "echo ' -> qemu-pbuilder %s'\"$1\"\n"
-	  "sleep 1s\n"
-	  "halt -f -p\n"	/* just halt myself if possible */
-	  "}\n",
-	  qemu_keyword);
-
+  write_second_stage_header(f, 0);
   fprintf(f,
 	  /* main code */
 	  "echo \n"
-	  "echo ' -> qemu-pbuilder second-stage' \n"
+	  "log.i 'qemu-pbuilder second-stage' \n"
 	  // Remove compatibility symlink
 	  "rm \"$BUILDDIR\"/pbuilder-run\n"
 	  //"mount -n /proc /proc -t proc\n" // this is done in first stage.
 	  "if [ \"${PBUILDER_INIT_VERSION:-0}\" -lt "XSTR(PBUILDER_INIT_VERSION)" ]; then\n"
-	  "    echo \"E: qemubuilder init script is out of date (${PBUILDER_INIT_VERSION:-0} < "XSTR(PBUILDER_INIT_VERSION)")\"\n"
-	  "    echo \"E: Please run qemubuilder --update\"\n"
+	  "    log.e \"qemubuilder init script is out of date (${PBUILDER_INIT_VERSION:-0} < "XSTR(PBUILDER_INIT_VERSION)")\"\n"
+	  "    log.e \"Please run qemubuilder --update\"\n"
 	  "    exit_from_qemu 1\n"
 	  "elif [ \"${PBUILDER_INIT_VERSION:-0}\" -gt "XSTR(PBUILDER_INIT_VERSION)" ]; then\n"
-	  "    echo \"E: qemubuilder init script is newer than expected (${PBUILDER_INIT_VERSION:-0} < "XSTR(PBUILDER_INIT_VERSION)")\"\n"
+	  "    log.e \"qemubuilder init script is newer than expected (${PBUILDER_INIT_VERSION:-0} < "XSTR(PBUILDER_INIT_VERSION)")\"\n"
 	  "    exit_from_qemu 1\n"
 	  "fi\n"
-	  "echo '  -> setting time to %s' \n"
+	  "log.i 'setting time to %s' \n"
 	  "date --set=\"%s\"\n"
-	  "echo '  -> configuring network' \n"
+	  "log.i 'configuring network' \n"
 	  "ifconfig -a\n"
 	  "export IFNAME=`/sbin/ifconfig -a | grep eth | head -n1 | awk '{print $1}'`\n"
 	  "dhclient $IFNAME\n"
@@ -819,17 +928,17 @@ static int run_second_stage_script
 		   "%s/pbuilder-run",
 		   pc->buildplace))
       {
-	fprintf(stderr,
-	        "E: Failed to allocate string for compatibility symlink path '%s/pbuilder-run': %s\n",
-		pc->buildplace,
-		strerror(errno));
+	log_printf(log_error,
+	    "Failed to allocate string for compatibility symlink path '%s/pbuilder-run': %s",
+	    pc->buildplace,
+	    strerror(errno));
 	goto out;
       }
     if (symlink("input/pbuilder-run", compat_symlink))
       {
-	fprintf(stderr,
-		"E: Failed to create compatibility symlink: %s\n",
-		strerror(errno));
+	log_printf(log_error,
+	    "Failed to create compatibility symlink: %s",
+	    strerror(errno));
 	free(compat_symlink);
 	goto out;
       }
@@ -856,7 +965,7 @@ static int run_second_stage_script
 
   if(hostcommand1)
     {
-      printf("running host command: %s\n", hostcommand1);
+      log_printf(log_info, "running host command: %s", hostcommand1);
       system(hostcommand1);
     }
 
@@ -878,7 +987,7 @@ static int run_second_stage_script
   /* commit the change here */
   if (save_result)
     {
-      printf(" -> commit change to qemu image\n");
+      log_printf(log_info, "committing changes to qemu image");
       ret=forkexeclp("qemu-img", "qemu-img",
 		     "commit",
 		     cowdevpath,
@@ -887,15 +996,15 @@ static int run_second_stage_script
 
   /* after-run */
   loop_mount(workblockdevicepath, pc->buildplace);
-  printf(" -> running post-run process\n");
+  log_printf(log_info, "running post-run process");
   if(hostcommand2)
     {
-      printf("running host command: %s\n", hostcommand2);
+      log_printf(log_info, "running host command: %s", hostcommand2);
       system(hostcommand2);
     }
   loop_umount(pc->buildplace);
   rmdir(pc->buildplace);
-  printf(" -> clean up COW device files\n");
+  log_printf(log_info, "clean up COW device files");
   unlink(workblockdevicepath);
   unlink(cowdevpath);
   ret=0;
@@ -1001,7 +1110,7 @@ int cpbuilder_create(const struct pbuilderconfig* pc)
     {
       /* could not create the buildplace here. */
       ret=1;
-      perror("mkdir");
+      log_perror("mkdir");
       goto out;
     }
 
@@ -1011,7 +1120,7 @@ int cpbuilder_create(const struct pbuilderconfig* pc)
       goto out;
     }
 
-  printf(" -> Invoking debootstrap\n");
+  log_printf(log_info, "Invoking debootstrap");
   ret=forkexeclp("debootstrap", "debootstrap",
 		 "--arch",
 		 pc->arch,
@@ -1022,8 +1131,7 @@ int cpbuilder_create(const struct pbuilderconfig* pc)
 		 NULL);
   if (ret)
     {
-      fprintf(stderr, "debootstrap failed with %i\n",
-	      ret);
+      log_printf(log_error, "debootstrap failed with %i", ret);
       goto umount_out;
     }
 
@@ -1036,7 +1144,7 @@ int cpbuilder_create(const struct pbuilderconfig* pc)
      we'll need think of other ways to work with them.
 
    */
-  printf(" -> Doing architecture-specific /dev population\n");
+  log_printf(log_info, "Doing architecture-specific /dev population");
 
   qemu_create_arch_devices(pc->buildplace, pc->arch);
 
@@ -1075,32 +1183,18 @@ int cpbuilder_create(const struct pbuilderconfig* pc)
   timestring=get_current_time_string();
 
   f = create_script(pc->buildplace, "input/pbuilder-run");
-  fprintf(f,
-	  "#!/bin/bash\n"
-	  /* define function to terminate qemu */
-	  "function exit_from_qemu() {\n"
-	  "%s\n"
-	  "sync\n"
-	  "sync\n"
-	  "sleep 1s\n"		/* sleep before sending dying message */
-	  "echo ' -> qemu-pbuilder %s$1'\n"
-	  "sleep 1s\n"
-	  "halt -f -p\n"	/* just halt myself if possible */
-	  "}\n",
-	  pc->debug?"echo \"Debug shell\"; /bin/bash":"",
-	  qemu_keyword);
-
+  write_second_stage_header(f, pc->debug);
   fprintf(f,
 	  /* start of main code */
 	  "export RET=0\n"
 	  "echo \n"
-	  "echo ' -> qemu-pbuilder second-stage' \n"
-	  "echo '  -> setting time to %s' \n"
+	  "log.i 'qemu-pbuilder second-stage' \n"
+	  "log.i 'setting time to %s' \n"
 	  "date --set=\"%s\"\n"
-	  "echo '  -> Running debootstrap second-stage script' \n"
+	  "log.i 'Running debootstrap second-stage script' \n"
 	  "touch /etc/udev/disabled\n" // work-around for #520742
 	  "/debootstrap/debootstrap --second-stage || ( "
-	  "  echo dumping debootstrap log\n"
+	  "  log.i \"dumping debootstrap log\"\n"
 	  "  cat /debootstrap/debootstrap.log\n"
 	  "  exit_from_qemu\n"
 	  "\n )\n"
@@ -1213,7 +1307,7 @@ int cpbuilder_build(const struct pbuilderconfig* pc, const char* dscfile)
 	   "ALLOWUNTRUSTED=%s /usr/lib/pbuilder/pbuilder-satisfydepends --control $BUILDDIR/*.dsc --internal-chrootexec 'chroot . ' %s \n"
 	   "apt-get install %s -y %s\n"
 	   "cd $BUILDDIR; /usr/bin/dpkg-source -x $(basename %s) \n"
-	   "echo ' -> Building the package'\n"
+	   "log.i 'Building the package'\n"
 	   EXECUTE_HOOKS("A")
 	   "if ! (\n"
 	   "    cd $BUILDDIR/*-*/\n"
@@ -1247,7 +1341,7 @@ int cpbuilder_build(const struct pbuilderconfig* pc, const char* dscfile)
 	   "        fi\n"
 	   "    done\n"
 	   "else\n"
-	   "    echo \"E: BUILDRESULT=[$BUILDRESULT] is not a directory.\"\n"
+	   "    log.e \"BUILDRESULT=[$BUILDRESULT] is not a directory.\"\n"
 	   "fi\n",
 	   pc->buildplace, pc->buildresult,
 	   pc->buildresultuid, pc->buildresultgid);
@@ -1288,7 +1382,7 @@ int cpbuilder_execute(const struct pbuilderconfig* pc, char** av)
   int ret;
 
   asprintf(&hostcommand,
-	   "[ -d %s/input ] || mkdir %s/input\n || echo \"E: Failed to create directory '%s/input'\"\n"
+	   "[ -d %s/input ] || mkdir %s/input\n || log.e \"Failed to create directory '%s/input'\"\n"
 	   "cp %s %s/input/runscript\n",
 	   pc->buildplace, pc->buildplace, pc->buildplace,
 	   av[0], pc->buildplace);
@@ -1319,28 +1413,26 @@ int cpbuilder_update(const struct pbuilderconfig* pc)
   ret = do_fsck(pc->basepath);
   if (ret)
     {
-      fprintf(stderr,
-	      "E: Failed fsck '%s'\n",
-	      pc->basepath);
+      log_printf(log_error, "Failed fsck '%s'", pc->basepath);
       return ret;
     }
 
   if (mkdir(pc->buildplace,0777))
     {
-      fprintf(stderr,
-	      "E: Could not create directory '%s': %s\n",
-	      pc->buildplace,
-	      strerror(errno));
+      log_printf(log_error,
+	  "Could not create directory '%s': %s",
+	  pc->buildplace,
+	  strerror(errno));
       return 1;
     }
 
   ret = loop_mount(pc->basepath, pc->buildplace);
   if (ret)
     {
-      fprintf(stderr,
-	      "E: Could not mount '%s' on '%s'\n",
-	      pc->basepath,
-	      pc->buildplace);
+      log_printf(log_error,
+	  "Could not mount '%s' on '%s'",
+	  pc->basepath,
+	  pc->buildplace);
       return ret;
     }
   FILE *g = create_script(pc->buildplace, "pbuilder-run");
@@ -1349,17 +1441,17 @@ int cpbuilder_update(const struct pbuilderconfig* pc)
   ret = loop_umount(pc->buildplace);
   if (ret)
     {
-      fprintf(stderr,
-	      "E: Could not unmount '%s'\n",
-	      pc->buildplace);
+      log_printf(log_error,
+	  "Could not unmount '%s'",
+	  pc->buildplace);
       return ret;
     }
   ret = rmdir(pc->buildplace);
   if (ret)
     {
-      fprintf(stderr,
-	      "E: Could not rmdir '%s'\n",
-	      pc->buildplace);
+      log_printf(log_error,
+	  "Could not rmdir '%s'",
+	  pc->buildplace);
       return ret;
     }
 
@@ -1381,7 +1473,7 @@ int cpbuilder_update(const struct pbuilderconfig* pc)
 		 , pc->allow_untrusted?"--force-yes":""
 		 , pc->extrapackages?pc->extrapackages:""))
     {
-      fprintf(stderr, "qemubuilder: out of memory.\n");
+      log_printf(log_error, "qemubuilder: out of memory.");
       return 1;
     }
 
